"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.resnet import ResNet50
from tf_keras.src.applications.resnet import decode_predictions
from tf_keras.src.applications.resnet import preprocess_input
