"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.activations import deserialize
from tf_keras.src.activations import elu
from tf_keras.src.activations import exponential
from tf_keras.src.activations import gelu
from tf_keras.src.activations import get
from tf_keras.src.activations import hard_sigmoid
from tf_keras.src.activations import linear
from tf_keras.src.activations import mish
from tf_keras.src.activations import relu
from tf_keras.src.activations import selu
from tf_keras.src.activations import serialize
from tf_keras.src.activations import sigmoid
from tf_keras.src.activations import softmax
from tf_keras.src.activations import softplus
from tf_keras.src.activations import softsign
from tf_keras.src.activations import swish
from tf_keras.src.activations import tanh
