"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras.datasets import boston_housing
from tf_keras.api._v1.keras.datasets import cifar10
from tf_keras.api._v1.keras.datasets import cifar100
from tf_keras.api._v1.keras.datasets import fashion_mnist
from tf_keras.api._v1.keras.datasets import imdb
from tf_keras.api._v1.keras.datasets import mnist
from tf_keras.api._v1.keras.datasets import reuters
