"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.datasets.reuters import get_label_names
from tf_keras.src.datasets.reuters import get_word_index
from tf_keras.src.datasets.reuters import load_data
