"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.mobilenet_v2 import MobileNetV2
from tf_keras.src.applications.mobilenet_v2 import decode_predictions
from tf_keras.src.applications.mobilenet_v2 import preprocess_input
