"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.optimizers.schedules.learning_rate_schedule import CosineDecay
from tf_keras.src.optimizers.schedules.learning_rate_schedule import CosineDecayRestarts
from tf_keras.src.optimizers.schedules.learning_rate_schedule import ExponentialDecay
from tf_keras.src.optimizers.schedules.learning_rate_schedule import InverseTimeDecay
from tf_keras.src.optimizers.schedules.learning_rate_schedule import LearningRateSchedule
from tf_keras.src.optimizers.schedules.learning_rate_schedule import PiecewiseConstantDecay
from tf_keras.src.optimizers.schedules.learning_rate_schedule import PolynomialDecay
from tf_keras.src.optimizers.schedules.learning_rate_schedule import deserialize
from tf_keras.src.optimizers.schedules.learning_rate_schedule import serialize
